//******************************************************************************************************************
// SerialRelay.ino
// Control GPIO connected Relays by PC USB Serial Port.         George Thomas, W5JDX 02/13/2023
// This code is released to the public domain.                  https://www.amateurlogic.tv
//******************************************************************************************************************
/*
  When new serial data arrives, this sketch adds it to a string.
  When a newline character is received, the loop parses the string and clears it.
  If the string contains "RELAY", the string is parsed to get any number
  immediately following.

  If a number is found and it's value is from 1 to 10, a digitalWrite
  command sends a LOW or HIGH to the GPIO port of the number. After the
  number include a space followed by either "ON" or "OFF" for the desired output.
  NOTE: relays are addressed 1 to 10 starting at GPIO pin 2. RELAY01 = GPIO pin 2, RELAY10 = GPIO pin 11, etc.
  example:
    RELAY1 ON    (sets pin 2 to LOW)
    RELAY1 OFF   (sets pin 2 to HIGH)

  If the string contains "STATUS", the status of GPIO pins 2 to 11,
  prefixed with "st=" is returned in a string to the serial port.
  example: If Relays 1 and 10 are on. (GPIO pins 2 and 11 are LOW)
    "st=1000000001"

  NOTE: Many commercial relay modules default to using a LOW to turn on the relays.
        Therefore, this program sends a LOW to turn on relays and a HIGH to turn them off.

  NOTE: By default, any time a serial program connects to the Arduino, the Arduino will automatically
        restart. This causes the state of all relays to reset. To prevent new
        serial connections from changing current relay states, connect a 10uf capacitor
        from the Arduino's Reset pin to ground.
        IMPORTANT: THIS CAPACITOR MUST BE DISCONNECTED TO UPLOAD A PROGRAM TO THE ARDUINO!

  NOTE: Companion SerialRelay.exe or PowerCycle.exe Windows programs may control this device if desired.
        You may also control this device directly from most any serial terminal program.
*/

// ** declare some global variables ************************************************************************************************
String inputString = "";        // a string to hold incoming data
String relayStringNum = "";     // parsed part of inputString
int relayNum = 0;               // number of a relay to control
int test = 0;                   // position in string of a found item
boolean stringComplete = false; // is inputString complete

// ** initialize the program at start up *******************************************************************************************
void setup() {
  //set pin mode
  for (int i = 2; i < 13; i++) {
    digitalWrite(i, HIGH);      // IMPORTANT: to avoid relays pulsing on when the unit is started, write HIGH to 
    pinMode(i, OUTPUT);         // each pin before setting its pinMode.
  }
  Serial.begin(9600);           // initialize serial port
  inputString.reserve(200);     // reserve 200 bytes to hold inputString
  //display this Bootup message on a connected terminal program
  Serial.println("Serial Relay Controller. (Arduino Uno  SerialRelay.ino");  // send this line to the serial terminal. Include a linefeed character.
  Serial.println("To turn ON or OFF a relay. (from 1 to 10)");
  Serial.println("    RELAY01 ON");
  Serial.println("    RELAY01 OFF");
  Serial.println("");
  Serial.println("To get the Status of all 10 relays");
  Serial.println("    STATUS");
}

// ** loop continuously during operation *******************************************************************************************
void loop() {
  // process the string when a newline character arrives:
  if (stringComplete) {
    inputString.toUpperCase();                // convert to upper case
    test = inputString.indexOf("RELAY");      // first character of this line is 0 position. -1 = not found.
    if (test != -1) {                         // check if 'RELAY' was received
      relayStringNum = inputString.substring(test + 5, inputString.length() - test);  //get everything after "RELAY"
      relayNum = relayStringNum.toInt();      // convert to integer to ignore everything else in the string except the number.
      if (relayNum > 0 || relayNum < 11) {    // if relay number is from 1 to 10
        test = inputString.indexOf("ON");     // check if 'ON' was received
        if (test != -1) {
          digitalWrite(relayNum + 1, LOW);   // turn GPIO pin on (LOW)
        }
        else {
          test = inputString.indexOf("OFF");  // check if 'OFF' was received
          if (test != -1) {
            digitalWrite(relayNum + 1, HIGH);  // turn GPIO pin OFF (HIGH)
          }
        }
      }
    }
    else {
      test = inputString.indexOf("STATUS");   // check if 'STATUS' was received
      if (test != -1) {                       // 'STATUS' was received
        String statusString = "";             // string to hold status bits
        int val = 0;
        int invertedVal = 0;
        for (int i = 2; i < 12; i++) {        // cycle through all 10 GPIO pins we are using
          val = digitalRead(i);
          if (val==1){                        // invert actual state of the GPIO pins to match relay logic
            invertedVal = 0;
          }
          else{
            invertedVal=1; 
         }         
          statusString = statusString + invertedVal;  // add inverted state of this bit to statusString
        }
        Serial.println("st=" + statusString); // display status in a connected terminal
      }
    }
    inputString = "";                         // clear inputString
    stringComplete = false;                   // clear stringComplete flag
  }
}

// ** SerialEvent occurs at the end of Loop whenever new data arrives in the arduino serial port. Multiple bytes of data may be available. ********
void serialEvent() {
  while (Serial.available()) {                // while there is data available in the serial port connection
    char inChar = (char)Serial.read();        // get the new byte
    inputString += inChar;                    // add it to inputString
    if (inChar == '\n') {                     // if newline character received
      inputString = inputString.substring(0, inputString.length() - 1); //strip off linefeed character
      stringComplete = true;                  // set a flag so the main loop knows we have a complete line
    }
  }

}
